/**
 * File: adc.c
 * Creator: Jonathan Klapel
 * Date: 29 May 2014
 * Description: ADC readings and writing to USB
 */ 

#include "adc.h"

/** Size of buffers from ADC */
#define BUFFER_LENGTH 512

/** Conversion values */
uint16_t g_adc_sample_data[10];
uint16_t audio_buf1[BUFFER_LENGTH];
uint16_t audio_buf2[BUFFER_LENGTH];
uint16_t audio_buf3[BUFFER_LENGTH];
uint8_t buffer1_full = 0;
uint8_t buffer2_full = 0;
uint8_t buffer3_full = 0;
uint32_t overload_count = 0;
uint8_t first_loop = 1;
uint32_t position = 0;

/** ADC instance */
struct adc_dev_inst g_adc_inst;

/* Time variables */
static uint32_t start_time;
static uint32_t end_time;
static uint32_t recording_duration;
static uint32_t playback_start;
static uint32_t time;
uint32_t sample_count = 0;
uint32_t actual_start_time = 0;
uint32_t actual_end_time = 0;
uint32_t end_time_offset = 0;

/* Audio file name */
char file_name[22];

/** ADC CDMA config value */
struct adc_cdma_first_config g_adc_cdma_first_cfg[1] = {
	{
		/* Select Vref for shift cycle */
		.zoomrange = ADC_ZOOMRANGE_0,
		/* Neg Input Channel 3 */
		.muxneg = ADC_MUXNEG_3,
		/* Pos Input Channel 2 */
		.muxpos = ADC_MUXPOS_2,
		/* Enables the internal voltage sources */
		.internal = ADC_INTERNAL_0,
		/* Disables the ADC gain error reduction */
		.gcomp = ADC_GCOMP_DIS,
		/* Disables the HWLA mode */
		.hwla = ADC_HWLA_EN,
		/* 12-bits resolution */
		.res = ADC_RES_12_BIT,
		/* Enables the differential mode */
		.bipolar = ADC_BIPOLAR_DIFFERENTIAL,
		/* Gain factor is 16x */
		.gain = ADC_GAIN_16X,
		.strig = 1,
		.tss = 1,
		.enstup = 1,
		.dw = 0
	}
};
/* PDCA Config values */
struct adc_pdca_config g_adc_pdca_cfg = {
	/* ADC Window Mode */
	.wm = false,
	/* ADC Nb channel */
	.nb_channels = 1,
	/* ADC Buffer */
	.buffer = g_adc_sample_data,
	/* ADC PDC Rx Channel */
	.pdc_rx_channel = CONFIG_ADC_PDCA_RX_CHANNEL,
	/* ADC PDC Tx Channel */
	.pdc_tx_channel = CONFIG_ADC_PDCA_TX_CHANNEL,
	/* ADC CDMA Configuration Structure */
	.cdma_cfg = (void *)&g_adc_cdma_first_cfg[0]
};

/**
 * Callback function for PDCA interrupt
 */
void pdca_transfer_done(enum pdca_channel_status status)
{
	/* Get PDCA RX channel status and check if PDCA transfer complete */
	if (status == PDCA_CH_TRANSFER_COMPLETED) {
		/* Time is before required start time, so pause for one moment */
		if (time < start_time) {
			actual_start_time = sw_timer_get_time();
			pdca_channel_write_load(CONFIG_ADC_PDCA_RX_CHANNEL, g_adc_sample_data, 1);
			first_loop = 1;
		/* Time is between start and end time, so data should be saved */
		} else if (start_time < time && (end_time+end_time_offset) > time) {
			/* Loop through three buffers filling up each */
			if (position == 0) {
				pdca_channel_write_load(CONFIG_ADC_PDCA_RX_CHANNEL, audio_buf2, BUFFER_LENGTH);
				buffer1_full = 1;
				position = 1;
			} else if (position == 1) {
				pdca_channel_write_load(CONFIG_ADC_PDCA_RX_CHANNEL, audio_buf3, BUFFER_LENGTH);
				buffer2_full = 1;
				position = 2;
			} else if (position == 2) {
				pdca_channel_write_load(CONFIG_ADC_PDCA_RX_CHANNEL, audio_buf1, BUFFER_LENGTH);
				buffer3_full = 1;
				position = 0;
			}
			
			/* Keep track of how many samples are supposed to be saved */
			sample_count += BUFFER_LENGTH;
			
			/* On first loop through, there are no useful values in buffers */
			if (first_loop == 1) {
				buffer1_full = 0;
				buffer2_full = 0;
				buffer3_full = 0;
				first_loop = 0;
				sample_count -= BUFFER_LENGTH;
			}
			
			/* Check if we are not writing fast enough, resulting in all buffers being filled */
			if(buffer1_full == 1 && buffer2_full == 1 && buffer3_full == 1) {
				overload_count++;
			}
		/* Time is past required end time so end ADC */
		} else if (actual_end_time == 0) {
			pdca_channel_write_load(CONFIG_ADC_PDCA_RX_CHANNEL, g_adc_sample_data, 10);
			
			/* Note actual end time */
			actual_end_time = sw_timer_get_time();
			
			/* Disable PDCA */
			pdca_channel_disable_interrupt(CONFIG_ADC_PDCA_RX_CHANNEL,PDCA_IER_TRC);
			pdca_channel_disable(CONFIG_ADC_PDCA_TX_CHANNEL);
			pdca_channel_disable(CONFIG_ADC_PDCA_RX_CHANNEL);
			
			/* Disable ADC */
			adc_stop_itimer(&g_adc_inst);
		}
	}
}

/**
 * ADC PDCA Set Config
 *
 * A rewrite of the base ADC PDCA setup in adcife.h.  This version
 * allows the ADC to record multiple values to the buffer before
 * interrupting the CPU.
 */
void adc_pdca_set_config2(struct adc_pdca_config *cfg)
{
	/* Enable PDCA module clock */
	pdca_enable(PDCA);

	/* PDCA channel options */
	pdca_channel_config_t PDCA_TX_CONFIGS = {
		/* memory address */
		.addr = (void *)cfg->cdma_cfg,
		/* select peripheral */
		.pid = ADCIFE_PDCA_ID_TX,
		/* transfer counter */
		.size = cfg->wm == true ? (cfg->nb_channels)*2 : cfg->nb_channels,
		/* next memory address */
		.r_addr = NULL,
		/* next transfer counter */
		.r_size = 0,
		/* select size of the transfer */
		.transfer_size = PDCA_MR_SIZE_WORD,
		/* set ring buffer */
		.ring = true
	};
	pdca_channel_config_t PDCA_RX_CONFIGS = {
		/* memory address */
		.addr = (void *)cfg->buffer,
		/* select peripheral */
		.pid = ADCIFE_PDCA_ID_RX,
		/* transfer counter */
		.size = 1,
		/* next memory address */
		.r_addr = NULL,
		/* next transfer counter */
		.r_size = 0,
		/* select size of the transfer */
		.transfer_size = PDCA_MR_SIZE_HALF_WORD
	};
	/* Init PDCA channel with the pdca_options. */
	pdca_channel_set_config(cfg->pdc_tx_channel, &PDCA_TX_CONFIGS);
	pdca_channel_set_config(cfg->pdc_rx_channel, &PDCA_RX_CONFIGS);
	/* Enable PDCA channel */
	pdca_channel_enable(cfg->pdc_tx_channel);
	pdca_channel_enable(cfg->pdc_rx_channel);
}

/**
 * Start ADC function
 *
 * Initialize ADC, set clock and timing, and set ADC to given mode.
 */
void start_adc(void)
{
	struct adc_config adc_cfg = {
		.prescal = ADC_PRESCAL_DIV4,
		.clksel = ADC_CLKSEL_APBCLK,
		.speed = ADC_SPEED_300K,
		/* ADC Reference voltage is 0.625*VCC */
		.refsel = ADC_REFSEL_1,
		.start_up = CONFIG_ADC_STARTUP
	};
	
	/* Calculate ADC internal timer value */
	uint32_t internal_timer_max_count = (sysclk_get_pba_hz())/((uint32_t)(4 << adc_cfg.prescal)*(uint32_t)WAVFILE_SAMPLES_PER_SECOND)-40;

	/* Initialize and enable ADC */
	if(adc_init(&g_adc_inst, ADCIFE, &adc_cfg) != STATUS_OK) {
		puts("-F- ADC Init Fail!\n\r");
		while(1);
	}
	if(adc_enable(&g_adc_inst) != STATUS_OK) {
		puts("-F- ADC Enable Fail!\n\r");
		while(1);
	}

	/* Configure PDCA */
	adc_disable_interrupt(&g_adc_inst, ADC_SEQ_SEOC);
	adc_pdca_set_config2(&g_adc_pdca_cfg);
	
	pdca_channel_set_callback(CONFIG_ADC_PDCA_RX_CHANNEL, pdca_transfer_done, PDCA_0_IRQn, 1, PDCA_IER_TRC);
	pdca_channel_write_load(CONFIG_ADC_PDCA_TX_CHANNEL, g_adc_cdma_first_cfg, 1);
	pdca_channel_write_reload(CONFIG_ADC_PDCA_TX_CHANNEL, g_adc_cdma_first_cfg, 1);

	adc_configure_trigger(&g_adc_inst, ADC_TRIG_INTL_TIMER);
	adc_configure_itimer_period(&g_adc_inst,internal_timer_max_count);
	adc_start_itimer(&g_adc_inst);

	adc_configure_gain(&g_adc_inst, ADC_GAIN_16X);
}

/**
 * ADC Task function
 *
 * OS task used for starting and running audio recordings.
 */
void adc_task(void) {
	printf("ADC Task Started\r\n");
	
	static FIL audio_file_object;
	actual_start_time = 0;
	actual_end_time = 0;
	char str_time[10];
	
	time = sw_timer_get_time();
	
	strcpy(file_name, "0:");
	strcat(file_name, "audio_");
	strncat(file_name,u32_itoa(time,str_time),10);
	strcat(file_name, ".wav");
	
	/* Open initial WAV file */
	wavfile_open(&audio_file_object, &file_name,WAVFILE_SAMPLES_PER_SECOND*5+BUFFER_LENGTH*3);
	
	/* Add buffer time to end of recording */
	end_time_offset = 1000*1000*BUFFER_LENGTH*3/WAVFILE_SAMPLES_PER_SECOND;
	
	/* Check to make sure we aren't late starting recording */
	if(time > start_time) {
		printf("Missed recording start time by %u ms\r\n",(time-start_time));
	}
	
	/* Make sure no one else interrupts us */
	taskENTER_CRITICAL();
	
	/* Start the ADC processing data from mic */
	start_adc();

	uint32_t bytes_written = 0;
	
	/* Write data to the USB while there is potentially data to be written */
	while ((end_time+end_time_offset) > time || buffer1_full == 1 || buffer2_full == 1 || buffer3_full == 1) {
		if (buffer1_full == 1) {
			uint16_t i = 0;
			while (i < BUFFER_LENGTH) {
				audio_buf1[i] ^= 0x8000;
				i++;
			}
			bytes_written +=  wavfile_write(&audio_file_object,audio_buf1,BUFFER_LENGTH);
			buffer1_full = 0;
		}
		if (buffer2_full == 1) {
			uint16_t i = 0;
			while (i < BUFFER_LENGTH) {
				audio_buf2[i] ^= 0x8000;
				i++;
			}
			bytes_written += wavfile_write(&audio_file_object,audio_buf2,BUFFER_LENGTH);
			buffer2_full = 0;
		}
		if (buffer3_full == 1) {
			uint16_t i = 0;
			while (i < BUFFER_LENGTH) {
				audio_buf3[i] ^= 0x8000;
				i++;
			}
			bytes_written += wavfile_write(&audio_file_object,audio_buf3,BUFFER_LENGTH);
			buffer3_full = 0;
		}
		time = sw_timer_get_time();
	}
	
	/* Finished with everything.  Close up and leave. */
	taskEXIT_CRITICAL();
	
	printf("Finished.  Overloaded %u times. Samples: %u\tBytes Written: %u\r\n",overload_count,sample_count,bytes_written);

	wavfile_close(&audio_file_object);
	
	vTaskSuspend(NULL);
	adc_recording_crop();
	
	printf("Exiting audio recording\r\n");
	vTaskDelete(NULL);
}

/**
 * Set Recording Times function
 *
 * Sets the start, end, and duration times based on what is received from the ground node
 */
void set_recording_times(uint32_t start, uint32_t end, uint32_t recording_length) {
	start_time = start;
	end_time = end;
	recording_duration = recording_length;
}

/**
 * Set Playback Start function
 *
 * Sets the time that the audio playback started on the ground node
 */
void set_playback_start(uint32_t actual_playback_start) {
	playback_start = actual_playback_start;
}

/**
 * ADC Recording Crop
 *
 * Removes excess audio at the beginning and ending of files
 */
void adc_recording_crop(void) {
	char file_name2[22];
	char str_time[10];

	strcpy(file_name2, "0:");
	strcat(file_name2, "edaud_");
	strncat(file_name2,u32_itoa((start_time+1),str_time),10);
	strcat(file_name2, ".wav");

	/* wavwriter library takes care of the actual cropping */
	wavfile_crop(&file_name, &file_name2,actual_start_time,playback_start,recording_duration);
}