/**
 * File: main.c
 * Creator: Jonathan Klapel
 * Date: 29 May 2014
 * Description: Remote (Quadcopter) Node Application
 *
 * APPLICATION USE
 * Setup:
 * - Insert Zigbit Wireless Module into EXT1
 * - Insert Flash Drive into SAM4L USB port
 * - Connect microphone assembly into pins PA07 (ADC+) and PB02 (ADC-) of EXT2
 * - Connect DEBUG USB port to PC and open console on COM port
 *
 * COM Settings:
 *  Speed: 38400 kbps
 *  Data bits: 8
 *  Stop bits: 1
 *  Parity: None
 *  Flow Control: None
 *
 * Use Instructions:
 *  Once power is supplied, the device will automatically start up
 *  operating system, setup WPAN network and wait for devices to
 *  connect.  Once connected, devices can instruct remote node to
 *  perform additional operations.
 */

#include <string.h>
#include "conf_board.h"
#include "conf_usb_host.h"
#include "avr2025_mac.h"
#include "led.h"
#include "delay.h"
#include "sleepmgr.h"
#include <asf.h>
#include "zigbit.h"
#include "adc.h"
#include "arm_math.h"

/* === DEFINES ========================= */
/* Audio test file write length */
#define SOUND_SAMPLES 0x200


/* === VARIABLES ======================= */
/* Task Handles (References) */
xTaskHandle xWpanHandle;
xTaskHandle xClockSyncHandle;
xTaskHandle xAudioSyncHandle;

/* Semaphore Handles */
xSemaphoreHandle xSyncPulse = NULL;
xSemaphoreHandle xAckPulse = NULL;
xSemaphoreHandle xConfirmPulse = NULL;
xSemaphoreHandle xAudioPulse = NULL;

/* MSDU Handles of WPAN requests */
uint8_t ack_pulse_handle = 0;
uint8_t audio_start_handle = 0;

/* Timestamps of synchronization pulses */
uint32_t sync_pulse_timestamp = 0;
uint32_t ack_pulse_timestamp = 0;

/* Pre-defined delay */
const portTickType xDelay = 1000 / portTICK_RATE_MS;

/* Audio Recording Variables */

uint32_t recording_start_time = 0;
uint32_t recording_length = 0;


/* === FUNCTIONS ======================= */

/**
 * \brief Called if stack overflow during execution
 */
extern void vApplicationStackOverflowHook(xTaskHandle *pxTask, signed char *pcTaskName)
{
	printf("stack overflow %x %s\r\n", pxTask, (portCHAR *)pcTaskName);
	for (;;) {
	}
}

/**
 * \brief This function is called by FreeRTOS idle task
 */
extern void vApplicationIdleHook(void)
{
}

/**
 * \brief This function is called by FreeRTOS each tick
 */
extern void vApplicationTickHook(void)
{
}

/**
 * \brief Configure the console UART.
 */
static void configure_console(void)
{
	const usart_serial_options_t uart_serial_options = {
		.baudrate = CONF_UART_BAUDRATE,
#if (defined CONF_UART_CHAR_LENGTH)
		.charlength = CONF_UART_CHAR_LENGTH,
#endif
		.paritytype = CONF_UART_PARITY,
#if (defined CONF_UART_STOP_BITS)
		.stopbits = CONF_UART_STOP_BITS,
#endif
	};

	/* Configure console UART. */
	stdio_serial_init(CONF_UART, &uart_serial_options);

	/* Specify that stdout should not be buffered. */
#if defined(__GNUC__)
	setbuf(stdout, NULL);
#else
	/* Already the case in IAR's Normal DLIB default configuration: printf()
	 * emits one character at a time.
	 */
#endif
}

/**
 * WPAN task
 * 
 * Start Time: On OS startup
 * End Time: On OS Termination
 * Function: Performs all wireless processing
 */
static void wpan_run_task(void) {
	printf("WPAN Task Started\r\n");
	while(1) {
		wpan_task();
	}
}

/**
 * AUDIO task
 *
 * Start Time: On OS Startup
 * End Time: On OS Termination
 * Function: Handles received audio recording requests
 * Note: Blocked on xAudioPulse semaphore
 */
static void audio_request_task(void) {
	vSemaphoreCreateBinary(xAudioPulse);
	if( xAudioPulse == NULL)
		printf("Audio semaphore not created properly\r\n");
	while (1) {
		if (xSemaphoreTake( xAudioPulse, portMAX_DELAY ) == pdTRUE && recording_start_time > 0) {
			if(	ui_identify_lun()) {
				printf("Running Audio Request\r\n");
				set_recording_times(recording_start_time, (recording_start_time + recording_length), recording_length);
				if (xTaskCreate(adc_task, "ADC", TASK_1024_STACK_SIZE*3, NULL, TASK_STACK_PRIORITY_0+2, &xAdcHandle) != pdPASS) {
					printf("Failed to create ADC task\r\n");
				}
			} else {
				printf("No USB device available for writing\r\n");
			}
		}
	}
}

/**
 * CLOCK SYNC task
 *
 * Start Time: On wireless connection established
 * End Time: On OS Termination
 * Function: Synchronizes with the ground node
 * Note: Blocked on xSyncPulse, xAckPulse, and xConfirmPulse
 */
static void clock_sync_task(void) {
	printf("Clock Sync Task Started\r\n");
	
	vSemaphoreCreateBinary(xSyncPulse);
	vSemaphoreCreateBinary(xAckPulse);
	vSemaphoreCreateBinary(xConfirmPulse);
	
	if( xSyncPulse == NULL || xAckPulse == NULL || xConfirmPulse == NULL) 
		printf("Clock semaphores not created properly\r\n");
		
	while(1) {
		// Ground node starts synchronization so await first sync pulse
		if (xSemaphoreTake( xSyncPulse, (portTickType)100 ) == pdTRUE) {
			// Send ACK pulse and await proper send confirmation
			ack_pulse_handle = send_message(ACK_PULSE,0,NULL);
			if (xSemaphoreTake(xAckPulse, (portTickType)100) == pdTRUE) {
				uint8_t payload[8];
				uint32_t temp_time = ack_pulse_timestamp;
				for (uint8_t i = 0; i < 8; i++) {
					payload[7-i] = temp_time % 256;
					temp_time = temp_time / 256;
					if (i == 3) {
						temp_time = sync_pulse_timestamp;
					}
				}
				// Send times T2 and T3 of TPSN to ground node
				send_message(TIMESTAMP_PULSE,8,&payload);
				if (xSemaphoreTake(xConfirmPulse,(portTickType)1000) == pdTRUE) {
					printf("Time confirmed\r\n");
				}
			} else {
				// Sending did not work properly so restart everything
				xSemaphoreTake(xConfirmPulse,(portTickType)0);
			}
		} else {
			// After sync pulse timeout, restart sync process, in case ground node is in different part of sync
			xSemaphoreTake(xAckPulse,(portTickType)0);
			xSemaphoreTake(xConfirmPulse,(portTickType)0);
		}
	}
}

/**
 * Test_File
 *
 * Writes a sample sine signal sweep to the USB drive.
 *
 * Not normally used but useful for creating ideal signal sweeps.
 */
void test_file(void) {
	printf("Generating Audio File\r\n");
	
	while(ui_identify_lun() == false) {;}
	
	float32_t duration = 2;
	float32_t fstart = 500;
	float32_t fend = 10000;
	
	static FIL audio_file_object;
	char file_name[22];
	
	float32_t inverse_sample_rate = 1/(float32_t)WAVFILE_SAMPLES_PER_SECOND;	
	float32_t pow_val = (1/((float32_t)WAVFILE_SAMPLES_PER_SECOND*duration));
	float32_t freq_change = (pow(fend,pow_val) - pow(fstart,pow_val))/pow(fstart,pow_val);
	
	uint32_t i = 0;
	uint32_t count;
	q15_t q15_sin_arg;
	int16_t samples_left[SOUND_SAMPLES];
	float32_t frequency = fstart;
	float32_t phase = 0;
	uint32_t loop_max = WAVFILE_SAMPLES_PER_SECOND*duration;
	
	strcpy(file_name, "0:");
	strcat(file_name, "genau_0000000001");
	strcat(file_name, ".wav");
	
	wavfile_open(&audio_file_object, &file_name,loop_max);

	while(i < loop_max) {
		count = 0;
		while (count < (SOUND_SAMPLES))	{
			frequency = frequency + frequency*freq_change;
			phase = modf(phase + frequency*inverse_sample_rate,NULL);
			arm_float_to_q15(&phase, &q15_sin_arg, 1);
			samples_left[count] = (int16_t)(arm_sin_q15(q15_sin_arg)) >> 1;
			count++;
		}
		wavfile_write(&audio_file_object, samples_left,count);
		
		i += count;
	}
	
	wavfile_close(&audio_file_object);
}

/**
 * Main function
 */
int main(void)
{
	/* Initialize interrupts */
	irq_initialize_vectors();
	
	/* Initialize clocks */
	sysclk_init();

	/* Initialize the board	 */
	board_init();

	/* Initiailize timers */
	sw_timer_init();
	
	/* Initialize the console uart */
	configure_console();

	/* Output app info infomation. */
	printf("-- REMOTE (QUADCOPTER) NODE --\n\r");
	printf("-- %s\n\r", BOARD_NAME);
	printf("-- Compiled: %s %s --\n\r", __DATE__, __TIME__);

	/* Initialize wireless */
	if (MAC_SUCCESS != wpan_init()) {
		printf("No MAC SUCCESS\r\n");
	}

	/* Initialize LEDs. */
	LED_On(LED_START);     /* indicating application is started */
	LED_Off(LED_NWK_SETUP); /* indicating node is associated */
	LED_Off(LED_DATA);     /* indicating successful data transmission */

	/* Initialize CPU IRQs*/
	cpu_irq_enable();

	/* Setup timer for Indicator LED */
	sw_timer_get_id(&TIMER_LED_OFF);
	
	// Start USB host stack
	uhc_start();
	
	/* Possibly generate test signal sweep file */
	//test_file();

	/* Set WPAN node status */
	node_status = COLD_START_RESET;

	/* Reset the MAC layer to the default values */
	wpan_mlme_reset_req(true);

	/* Create WPAN Task */
	if (xTaskCreate(wpan_run_task, "WPAN", TASK_2048_STACK_SIZE, NULL, TASK_STACK_PRIORITY_0, &xWpanHandle) != pdPASS) {
		printf("Failed to create WPAN task\r\n");
	}

	/* Start the OS scheduler. */
	vTaskStartScheduler();

	/* Will only get here if there was insufficient memory to create the idle task. */
	return 0;
}

/*
 * Wireless Ready function
 *
 * Runs processes that can only be started after wireless is connected
 * Includes setting up CLOCK SYNC and AUDIO tasks
 */
void wireless_ready(void) {
	printf("Wireless Ready\r\n");
	if (xClockSyncHandle == NULL) {
		if (xTaskCreate(clock_sync_task, "CLOCKSYNC", TASK_2048_STACK_SIZE, NULL, TASK_STACK_PRIORITY_0, &xClockSyncHandle) != pdPASS) {
			printf("Failed to create Clock Sync task\r\n");
		}
		if (xTaskCreate(audio_request_task, "AUDIOREQUEST", TASK_2048_STACK_SIZE, NULL, TASK_STACK_PRIORITY_0+1, &xAudioSyncHandle) != pdPASS) {
			printf("Failed to create Audio Sync task\r\n");
		}
	}
}

/**
 * Data Received function
 *
 * Handles any data that is received from the Zigbit module
 */
void data_received(uint8_t msduLength,uint8_t *msdu,uint32_t Timestamp) {
	static signed portBASE_TYPE xHigherPriorityTaskWoken;
	uint32_t actual_playback_start = 0;
	/* Perform different actions based on what sort of message was received */
	switch (msdu[0]) {
		case UNKNOWN:
			printf("Unknown data received\r\n");
			break;
		case SYNC_PULSE:
			sync_pulse_timestamp = Timestamp;
			xSemaphoreGiveFromISR( xSyncPulse, &xHigherPriorityTaskWoken );
			break;
		case CONFIRM_SYNC:
			xSemaphoreGiveFromISR( xConfirmPulse, &xHigherPriorityTaskWoken );
			break;
		case START_RECORDING:
			recording_start_time = ((msdu[1]*256+msdu[2])*256+msdu[3])*256+msdu[4];
			recording_length = ((msdu[5]*256+msdu[6])*256+msdu[7])*256+msdu[8];
			xSemaphoreGiveFromISR( xAudioPulse, &xHigherPriorityTaskWoken );
			break;
		case RECORDING_STATUS:
			send_message(RECORDING_STATUS,0,NULL);
			actual_playback_start = ((msdu[1]*256+msdu[2])*256+msdu[3])*256+msdu[4];
			set_playback_start(actual_playback_start);
			vTaskPrioritySet(xAdcHandle,TASK_STACK_PRIORITY_0);
			vTaskResume(xAdcHandle);
			break;
		default:
			printf("Bad Message Received\r\n");
			break;
	}
}

/**
 * Data Transmitted function
 *
 * Called by the WPAN task upon successful completion of a
 * sent message.  Other tasks may need to know when a message
 * has been properly sent.
 */
void data_transmitted(uint8_t msduHandle, uint8_t status, uint32_t Timestamp) {
	static signed portBASE_TYPE xHigherPriorityTaskWoken;
	if (msduHandle == ack_pulse_handle)	{
		ack_pulse_timestamp = Timestamp;
		xSemaphoreGiveFromISR( xAckPulse, &xHigherPriorityTaskWoken );
	}
	if (msduHandle == audio_start_handle) {
		printf("Transmitted\r\n");
		vTaskResume(xAudioSyncHandle);
	}
}