/**
 * File: ui.c
 * Creator: Jonathan Klapel
 * Date: 08 June 2014
 * Description: Functions required for USB thumb drive
 */ 

#include "conf_usb_host.h"
#include <asf.h>
#include <string.h>
#include "led.h"

/* === USB Variables =========================================================== */
static uhc_enum_status_t ui_enum_status = UHC_ENUM_DISCONNECT;

#define MAX_DRIVE      _VOLUMES
#define TEST_FILE_NAME "0:uhi_msc_test.txt"
#define MSG_TEST "Test UHI MSC\n"

static FATFS fs;
static FIL file_object;

static uhc_enum_status_t main_enum_status = UHC_ENUM_DISCONNECT;

static char test_file_name[] = { TEST_FILE_NAME };

static uint16_t ui_usb_sof_counter = 0;

static bool need_lun = true;
static uint8_t connected_lun = 0;

/* === Functions =========================================================== */

/**
 * UI USB SOF Event Function
 *
 * Notify that a SOF has been sent (each 1 ms)
 */
void ui_usb_sof_event(void) {
	ui_usb_sof_counter++;
}

/** 
 * UI USB Connection Event Function
 *
 * Notify that a USB device has been connected or disconnected.
 *
 * Parameter: dev         Pointer on USB device information
 * Parameter: b_present   true, if the device has been connected
 */
void ui_usb_connection_event(uhc_device_t * dev, bool b_present) {
	need_lun = true;
	ui_usb_sof_counter = 0;
	if (!b_present) {
		printf("USB Disconnected\r\n");
	} else {
		printf("USB Connected\r\n");
	}
}

/**
 * UI Identify LUN function
 *
 * Identifies the LUN of the connected device
 */
bool ui_identify_lun(void) {
	if (ui_usb_sof_counter > 2000 && need_lun) {
		ui_usb_sof_counter = 0;
		volatile uint8_t lun;
		FRESULT res;

		for (lun = LUN_ID_USB; (lun < LUN_ID_USB + uhi_msc_mem_get_lun()) &&
				(lun < MAX_DRIVE); lun++) {
			
			/* Mount drive */
			memset(&fs, 0, sizeof(FATFS));
			res = f_mount(lun, &fs);
			if (FR_INVALID_DRIVE == res) {
				continue;
			}
			
			/* Create a test file on the disk */
			test_file_name[0] = lun + '0';
			uint32_t time = sw_timer_get_time();
			char file_name[21];
			char str_time[10];
			
			strcpy(file_name, "0:");
			strcat(file_name, "test_");
			strncat(file_name,u32_itoa(sw_timer_get_time(),str_time),10);
			strcat(file_name, ".txt");
			
			res = f_open(&file_object,(char const *)file_name,FA_CREATE_ALWAYS | FA_WRITE);
			
			if (res != FR_OK) {
				// LUN test error
				printf("FR not OK\r\n");
				f_close(&file_object);
				continue;
			}
			
			/* Write to test file */
			f_puts(MSG_TEST, &file_object);
			f_printf(&file_object, "Lun: %u\n",lun);
			
			// LUN test OK
			f_close(&file_object);
			
			need_lun = false;
			connected_lun = lun;
		}
	}
	return !need_lun;
}

/**
 * Get Connected LUN Function
 *
 * Gets the details of a connected LUN
 */
uint8_t get_connected_lun(void) {
	return connected_lun;
}

/**
 * Callbacks to show the MSC read and write access
 */
void ui_start_read(void)
{
	LED_On(LED0);
}

void ui_stop_read(void)
{
	LED_Off(LED0);
}

void ui_start_write(void)
{
	LED_On(LED0);
}

void ui_stop_write(void)
{
	LED_Off(LED0);
}

