/**
 * File: wavwriter.c
 * Creator: Jonathan Klapel
 * Date: 08 June 2014
 * Description: Writes WAV files
 */ 

#include "wavwriter.h"
#include <string.h>

/* WAV file header */
struct wavfile_header {
	char	riff_tag[4];
	int		riff_length;
	char	wave_tag[4];
	char	fmt_tag[4];
	int		fmt_length;
	short	audio_format;
	short	num_channels;
	int		sample_rate;
	int		byte_rate;
	short	block_align;
	short	bits_per_sample;
	char	data_tag[4];
	int		data_length;
};

/**
 * WAV File Open function
 *
 * Begins a WAV file that will have audio data streamed to it.
 */
void wavfile_open(FIL *file_object, char const *file_name, uint32_t length )
{
	/* Format header of WAV file*/
	struct wavfile_header header;

	int samples_per_second = WAVFILE_SAMPLES_PER_SECOND;
	int bits_per_sample = 16;

	strncpy(header.riff_tag,"RIFF",4);
	strncpy(header.wave_tag,"WAVE",4);
	strncpy(header.fmt_tag,"fmt ",4);
	strncpy(header.data_tag,"data",4);

	header.riff_length = 0;
	header.fmt_length = 16;
	header.audio_format = 1;
	header.num_channels = 1;
	header.sample_rate = samples_per_second;
	header.byte_rate = samples_per_second*(bits_per_sample/8);
	header.block_align = bits_per_sample/8;
	header.bits_per_sample = bits_per_sample;
	header.data_length = 0;
	
	FRESULT res;
	
	/* Create and open file for editing */
	res = f_open(file_object,(char const *)file_name,FA_CREATE_ALWAYS | FA_WRITE);
	if (res != FR_OK) {
		printf("Could not open file object\r\n");
		f_close(file_object);
		return;
	}
	
	/* Write header to file */
	uint32_t bytes_written;
	while( f_write(file_object,&header,sizeof(header),&bytes_written) != FR_OK ) {;}
	
	/* Pre-allocate space for remained of file */
	f_lseek(file_object,length*2);
	f_lseek(file_object,sizeof(struct wavfile_header));
}

/**
 * WAV File Write function
 *
 * Writes an array of bytes to the file
 */
uint32_t wavfile_write(FIL *file_object, short data[], int length )
{
	/* Write bytes to the file */
	uint32_t bytes_written;
	while( f_write(file_object,data,sizeof(short)*length,&bytes_written) != FR_OK) {;}
	
	return bytes_written;
}

/**
 * WAV File Close function
 *
 * Closes the WAV file once recording is finished
 */
void wavfile_close(FIL *file_object )
{
	uint32_t bytes_written;
	
	/* Remove any additional pre-allocated space that may exist */
	f_truncate(file_object);
	
	/* Get length of data sections to write into WAV file header chunks */
	int file_length = f_tell(file_object);
	int data_length = file_length - sizeof(struct wavfile_header);
	f_lseek(file_object,sizeof(struct wavfile_header) - sizeof(int));
	while( f_write(file_object,&data_length,sizeof(data_length),&bytes_written) != FR_OK ) {;}

	/* Get file length to write in WAV file header */
	int riff_length = file_length - 8;
	f_lseek(file_object,4);
	while( f_write(file_object,&riff_length,sizeof(riff_length),&bytes_written) != FR_OK ) {;}

	/* Close the file.  We're done with it. */
	f_close(file_object);
}

/**
 * WAV File Crop function
 *
 * Cuts the excess audio data off the front and back of the audio recording
 */
void wavfile_crop(char const *file1, char const *file2, uint32_t recording_start_time, uint32_t playback_start_time, uint32_t duration) {
	FRESULT res;
	FIL f_file1, f_file2;
	BYTE buffer[128];
	uint32_t br, bw;
	
	/* Open file1 for reading from */
	res = f_open(&f_file1, (char const *)file1,FA_OPEN_EXISTING | FA_READ);
	if (res) return (int)res;
	
	/* Open file2 for writing to */
	res = f_open(&f_file2, (char const *)file2,FA_CREATE_ALWAYS | FA_WRITE);
	if (res) return (int)res;
	
	/* Copy header of old file into new file */
	f_read(&f_file1, buffer, sizeof(struct wavfile_header), &br);
	if (res || br == 0) return; /* error or eof */
	res = f_write(&f_file2, buffer, br, &bw);            /* Write it to the destination file */
	if (res || bw < br) return; /* error or disk full */
	
	/* If there is excess audio on the front, skip over it.  If there is not enough audio, insert blank audio */
	if (recording_start_time < playback_start_time) {
		uint32_t remove_samples = (playback_start_time - recording_start_time) * WAVFILE_SAMPLES_PER_SECOND / (1000 * 1000) - CODE_OFFSET;
		remove_samples *= 2;
		res = f_lseek(&f_file1,f_tell(&f_file1)+remove_samples);
		if (res) return;
	} else {
		uint32_t add_samples = (recording_start_time - playback_start_time) * WAVFILE_SAMPLES_PER_SECOND / (1000 * 1000);
		while (add_samples > 0) {
			f_putc((TCHAR)0,&f_file2);
			f_putc((TCHAR)0,&f_file2);
			add_samples--;
		}
	}
	
	uint32_t i = 0;
	uint32_t max_samples = duration / (1000*1000) * (WAVFILE_SAMPLES_PER_SECOND*2) ;
	/* Copy source file to destination file, up to max audio duration*/
	while (i < max_samples) {
		if ((max_samples - i) < sizeof buffer) {
			res = f_read(&f_file1, buffer, (max_samples - i), &br);  /* Read a chunk of source file */
		} else {
			res = f_read(&f_file1, buffer, sizeof buffer, &br);  /* Read a chunk of source file */
		}
		if (res || br == 0) break; /* error or eof */
		res = f_write(&f_file2, buffer, br, &bw);            /* Write it to the destination file */
		if (res || bw < br) break; /* error or disk full */
		i += br;
	}
	
	/* Close open files */
	f_close(&f_file1);
	wavfile_close(&f_file2);
}
