/**
 * File: dac.c
 * Creator: Jonathan Klapel
 * Date: 08 June 2014
 * Description: DAC Audio output functions
 */ 

#include "dac.h"

/* Defines */
#define PDCA_ABDAC_CHANNEL0  0
#define PDCA_ABDAC_CHANNEL1  1
#define SAMPLE_RATE	 44100
#define RECORD_INDICTATOR	  PIN_PC08
#define SOUND_SAMPLES 0x100

/* Output buffer */
int16_t samples_buffer[SOUND_SAMPLES];

/* PDCA channel options */
static const pdca_channel_config_t pdca_abdac_config0 = {
	.addr = 0,   /* memory address */
	.pid = ABDACB_PDCA_ID_TX_CH0,     /* select peripheral */
	.size = 0,   /* transfer counter */
	.r_addr = 0,                   /* next memory address */
	.r_size = 0,                   /* next transfer counter */
	.ring = false,                 /* disable ring buffer mode */
	.transfer_size = PDCA_MR_SIZE_HALF_WORD /* select size of the transfer */
};
static const pdca_channel_config_t pdca_abdac_config1 = {
	.addr = 0,   /* memory address */
	.pid = ABDACB_PDCA_ID_TX_CH1,     /* select peripheral */
	.size = 0,   /* transfer counter */
	.r_addr = 0,                   /* next memory address */
	.r_size = 0,                   /* next transfer counter */
	.ring = false,                 /* disable ring buffer mode */
	.transfer_size = PDCA_MR_SIZE_HALF_WORD /* select size of the transfer */
};

/** ABDAC instance */
struct abdac_dev_inst g_abdac_inst;

/** ABDAC configuration */
struct abdac_config   g_abdac_cfg;

/* Output variables */
double fstart;
double fend;
double duration;
double record_duration;

uint32_t underrun = 0;

uint8_t audio_start_handle = 0;

/**
 * Audio Init Function
 *
 * Initializes the ABDAC and IO pin
 */
void audio_init(void) {
	/* Setup ABDAC */
	status_code_t status;
	
	/* Setup ADC */
	abdac_get_config_defaults(&g_abdac_cfg);
	g_abdac_cfg.sample_rate_hz = ABDAC_SAMPLE_RATE_44100;
	g_abdac_cfg.data_word_format = ABDAC_DATE_16BIT;
	g_abdac_cfg.mono = true;
	g_abdac_cfg.cmoc = false;
	status = abdac_init(&g_abdac_inst, ABDACB, &g_abdac_cfg);
	if (status != STATUS_OK) {
		printf("-- Initialization fails! --\r\n");
		while (1) {
		}
	}
	
	/* Setup IO Testing Pin */
	ioport_set_pin_dir(RECORD_INDICTATOR,IOPORT_DIR_OUTPUT);
	ioport_set_pin_level(RECORD_INDICTATOR,false);
}

/**
 * Audio Underrun Function
 *
 * Called when the DAC does not receive data fast enough
 */
void audio_underrun(void) {
	underrun++;
	abdac_clear_interrupt_flag(&g_abdac_inst, ABDAC_INTERRUPT_TXUR);
}

/**
 * Audio Task Function
 *
 * Primary function for performing audio measurements
 */
void audio_task(void) {
	printf("Audio Recording Task Started\r\n");
	uint8_t key;
	uint32_t i, count;
	uint8_t payload[8];

	printf("\tGenerated Signal Sweep: %u-%u over %u seconds\r\n",(uint32_t)fstart,(uint32_t)fend,(uint32_t)duration);
	printf("\tAudio recording duration: %u\r\n",(uint32_t)record_duration);
	
	/* Prep ABDAC */
	abdac_enable(&g_abdac_inst);
	abdac_clear_interrupt_flag(&g_abdac_inst, ABDAC_INTERRUPT_TXRDY);
	abdac_clear_interrupt_flag(&g_abdac_inst, ABDAC_INTERRUPT_TXUR);
	abdac_set_callback(&g_abdac_inst,ABDAC_INTERRUPT_TXUR,audio_underrun,1);

	/* Config PDCA module */
	pdca_enable(PDCA);
	pdca_channel_set_config(PDCA_ABDAC_CHANNEL0, &pdca_abdac_config0);
	pdca_channel_enable(PDCA_ABDAC_CHANNEL0);
	
	/* Set Volumes */
	abdac_set_volume0(&g_abdac_inst, false, 0x7FFF);
	abdac_set_volume1(&g_abdac_inst, false, 0x7FFF);
	
	/* Initial calculations */
	float_t inverse_sample_rate = 1/(float_t)SAMPLE_RATE;
	float32_t pow_val = (1/((float32_t)SAMPLE_RATE*duration));
	float32_t freq_change = (pow(fend,pow_val) - pow(fstart,pow_val))/pow(fstart,pow_val);
	
	/* Calculate recording times and send remote node notice to record */
	uint32_t start_time = get_global_time() + RECORD_DELAY_SECONDS * 1000 * 1000;
	uint32_t record_time = record_duration * 1000 * 1000;
	for (i = 0; i < 8; i++) {
		payload[7-i] = record_time % 256;
		record_time = record_time / 256;
		if (i == 3) {
			record_time = start_time;
		}
	}
	record_time = record_duration * 1000 * 1000;
	start_time += 1000;
	
	audio_start_handle = send_message(START_RECORDING,8,&payload);
	printf("Audio Handle: %u\r\n",audio_start_handle);
	
	/* Wait until message has been properly sent */
	vTaskSuspend(NULL);
	printf("Message Sent\r\n");
	
	i = 0;
	q15_t q15_sin_arg;
	float32_t frequency = fstart;
	float32_t phase = 0;
	uint32_t loop_max = SAMPLE_RATE*duration;
	
	taskENTER_CRITICAL();
	underrun = 0;
	count = 0;
	
	/* Prepare first samples */
	while (count < (SOUND_SAMPLES))	{
		frequency = frequency + frequency*freq_change;
		phase = modf(phase + frequency*inverse_sample_rate,NULL);
		arm_float_to_q15(&phase, &q15_sin_arg, 1);
		samples_buffer[count] = (int16_t)(arm_sin_q15(q15_sin_arg)^0x8000);
		count++;
	}
	
	/* Wait until it is time to actually start */
	uint32_t system_time = get_global_time();
	if (system_time > start_time) {
		printf("Missed start deadline\r\n");
	}
	while (system_time < start_time) {
		system_time = get_global_time();
	}
	
	pdca_channel_write_reload(PDCA_ABDAC_CHANNEL0,(void *)samples_buffer, SOUND_SAMPLES);
	ioport_set_pin_level(RECORD_INDICTATOR,true);
	uint32_t playback_start_time = get_global_time();
	underrun = 0;
	i += count;
	while(!(pdca_get_channel_status(PDCA_ABDAC_CHANNEL0)== PDCA_CH_COUNTER_RELOAD_IS_ZERO));
	
	/* Keep feeding samples to the PDCA/DAC */
	while(i < loop_max) {
		count = 0;
		while (count < (SOUND_SAMPLES))	{
			frequency = frequency + frequency*freq_change;
			phase = modf(phase + frequency*inverse_sample_rate,NULL);
			arm_float_to_q15(&phase, &q15_sin_arg, 1);
			samples_buffer[count] = (int16_t)(arm_sin_q15(q15_sin_arg)^0x8000);
			count++;
		}
		pdca_channel_write_reload(PDCA_ABDAC_CHANNEL0,(void *)samples_buffer, SOUND_SAMPLES);
		i += count;
		while(!(pdca_get_channel_status(PDCA_ABDAC_CHANNEL0)== PDCA_CH_COUNTER_RELOAD_IS_ZERO));
	}
	
	/* Once finished, close everything out */
	ioport_set_pin_level(RECORD_INDICTATOR,false);
	printf("Exiting the audio output with %u underruns\r\n\r\n",underrun);
	abdac_set_volume0(&g_abdac_inst, true, 0x7FFF);
	abdac_set_volume1(&g_abdac_inst, true, 0x7FFF);
	taskEXIT_CRITICAL();
	
	system_time = get_global_time();
	while (system_time < (start_time+record_time)) {
		system_time = get_global_time();
	}
	
	/* Send notice to remote node of audio start time */
	printf("Requesting Status\r\n");
	uint8_t msg[4];
	for (i = 0; i < 4; i++) {
		msg[3-i] = playback_start_time % 256;
		playback_start_time = playback_start_time / 256;
	}
	send_message(RECORDING_STATUS,4,&msg);
	vTaskSuspend(NULL);
	
	/* Delete task */
	printf("Recording Status found\r\n");
	vTaskDelete(NULL);
}

/**
 * Set Playback Values Function
 *
 * Sets the values for the next audio output and recording
 */
void set_playback_values(double freq_start, double freq_end, double play_duration, double rec_duration) {
	fstart = freq_start;
	fend = freq_end;
	duration = play_duration;
	record_duration = rec_duration;
}

/**
 * Get Audio Start Handle Function
 *
 * Returns the most recent msdue audio start handle
 */
uint8_t get_audio_start_handle(void) {
	return audio_start_handle;
}