/**
 * File: main.c
 * Creator: Jonathan Klapel
 * Date: 08 June 2014
 * Description: Ground Node Application
 *
 * APPLICATION USE
 * Setup:
 * - Insert Zigbit Wireless Module into EXT1
 * - Insert Flash Drive into SAM4L USB port
 * - Connect loudspeaker assembly into pin PC09 and GND
 * - Connect DEBUG USB port to PC and open console on COM port
 * - Or connect SAM4L USB port to PC and open USB application
 *
 * COM Settings:
 *  Speed: 38400 kbps
 *  Data bits: 8
 *  Stop bits: 1
 *  Parity: None
 *  Flow Control: None
 *
 * Use Instructions:
 *  Once power is supplied, the device will automatically start up
 *  operating system, start console menu, and try to connect to
 *  remote node.  Once connected, it can can command remote node to
 *  perform additional operations.
 */

#include <asf.h>
#include <string.h>
#include "conf_board.h"
#include "conf_usb.h"
#include "ui.h"
#include "string.h"
#include "led.h"
#include "delay.h"
#include "common_sw_timer.h"
#include "sleepmgr.h"
#include "FreeRTOS.h"
#include "queue.h"

/* OS Task Handles */
xTaskHandle xWpanHandle;
xTaskHandle xConsoleHandle;
xTaskHandle xUSBHandle;
xTaskHandle xClockSyncHandle;

/* Semaphore Handles */
xSemaphoreHandle xAckPulse = NULL;
xSemaphoreHandle xTimestampPulse = NULL;

/* MSDU Handle for Sync Pulse */
uint8_t msduHandle_sync = 0;

/* Standard RTOS Delay */
const portTickType xDelay = 100 / portTICK_RATE_MS;


/* Push Button Variables */
uint8_t DEBOUNCE;
#define PUSH_BUTTON_INT    PIN_PA06


/**
 * Application Stack Overflow Hook Function
 *
 * Called if stack overflow during execution
 */
extern void vApplicationStackOverflowHook(xTaskHandle *pxTask,
		signed char *pcTaskName)
{
	printf("stack overflow %x %s\r\n", pxTask, (portCHAR *)pcTaskName);
	for (;;) {
	}
}

/**
 * Application Idle Hook Function
 *
 * This function is called by FreeRTOS idle task
 */
extern void vApplicationIdleHook(void)
{
}

/**
 * Application Tick Hook Function
 *
 * This function is called by FreeRTOS each tick
 */
extern void vApplicationTickHook(void)
{
}

/**
 * Low Speed USB Task Function
 *
 * Performs tasks that high-speed USB takes care of automatically
 */
#ifdef   USB_DEVICE_LOW_SPEED
static void low_speed_usb_task(void) {	
	while (true) {
		// No USB "Keep a live" interrupt available in low speed
		// to scan generic interface then use main loop
		if (main_b_generic_enable) {
			static volatile uint16_t virtual_sof_sub = 0;
			static uint16_t virtual_sof = 0;
			if (sysclk_get_cpu_hz()/50000 == virtual_sof_sub++) {
				virtual_sof_sub = 0;
				static uint16_t virtual_sof = 0;
				ui_process(virtual_sof++);
			}
		}
	}
}
#endif

/**
 * Clock Sync Task Function
 *
 * Takes care of synchronizing the clocks between the nodes
 */
static void clock_sync_task(void) {
	printf("Starting Clock Sync Task\r\n");
	
	vSemaphoreCreateBinary(xAckPulse);
	vSemaphoreCreateBinary(xTimestampPulse);
	if (xTimestampPulse == NULL || xAckPulse == NULL)
		printf("Clock semaphores not created properly\r\n");
		
	while (1) {
		// Send Synchronization Pulse
		msduHandle_sync = send_message(SYNC_PULSE,0,NULL);
		// Wait for acknowledgment pulse
		if(xSemaphoreTake( xAckPulse, (portTickType)10 ) == pdTRUE) {
			// Wait for timestamp pulse
			if(xSemaphoreTake( xTimestampPulse, (portTickType)10 ) == pdTRUE) {
				// Perform final TPSN calculations
				calculate_offsets();
				// Send message confirming synchronization reception
				send_message(CONFIRM_SYNC,0,NULL);
				// Set how long to wait for next sync attempt based on sync status
				if (isTimeSynced() == 1) {
					vTaskDelay(50*xDelay);
				} else {
					vTaskDelay(15*xDelay);
				}
			} else {
				printf("Missed second sync request\r\n");
				vTaskDelay(15*xDelay);
			}
		} else {
			printf("Missed sync request\r\n");
			xSemaphoreTake( xTimestampPulse, (portTickType)0 );
			vTaskDelay(15*xDelay);
		}
	}
}

/**
 * WPAN Run Task Function
 *
 * Continues to perform wireless background tasks
 */
static void wpan_run_task(void) {
	printf("WPAN Task Started\r\n");
	while(1) {
		wpan_task();
		vTaskDelay( xDelay );
	}
}

/**
 * Get Int Function
 *
 * Retrieves an int entered at the UART console
 */
static uint32_t getint(void) {
	uint8_t key;
	uint32_t val = 0;
	key = getchar();
	putchar(key);
	key = key - '0';
	while (key < 10) {
		val = val*10 + key;
		key = getchar();
		putchar(key);
		key = key - '0';
	}
	printf("\r\n");
	return val;
}

/**
 * Audio Prep Function
 *
 * Upon selection "Audio Request" from the menu, this function
 *  is called to get more details about the timing and
 *  frequencies.
 */
static void audio_prep(void) {
	uint32_t num = 0;
	double fstart = 0;
	double fend = 0;
	double duration = 0;
	double record_duration = 0;
	
	printf("Enter the start frequency (100-20000): ");
	num = getint();
	fstart = (double)num;
		
	printf("Enter the end frequency (100-20000): ");
	num = getint();
	fend = (double)num;
	
	printf("Enter the duration of the signal sweep (1-10): ");
	num = getint();
	duration = (double)num;
	
	printf("Enter the duration of the audio recording (1-10): ");
	num = getint();
	record_duration = (double)num;
	
	set_playback_values(fstart,fend,duration,record_duration);
}

/** 
 * Display Menu Function
 */
static void display_menu(void) {
	printf("==== Menu Choices ====\r\n");
	printf("s: Check synchronization between devices\r\n");
	printf("a: Do audio recording\r\n");
	printf("m: Display menu\r\n");
}

/**
 * Console Task Function
 *
 * Main task for user interaction via UART
 */
static void console_task(void) {
	printf("Console Task Started\r\n");
	
	display_menu();
	uint8_t key;
	
	// Keep reading from the console
	while (1) {
		key = getchar();
		switch (key) {
		case 's':
		case 'S':
			// Status request
			printf("Sync Status:\r\n");
			printf("\tNode Status: %u\r\n",node_status);
			printf("\tTime Sync Status: %u\r\n",isTimeSynced());
			printf("\tDelta Change: %f \r\n",get_delta_change());
			display_menu();
			break;
		case 'a':
		case 'A':
			// Audio request
			audio_prep();
			if (isTimeSynced() == 1) {
				if (xTaskCreate(audio_task, "AUDIO", TASK_1024_STACK_SIZE*3, NULL, TASK_STACK_PRIORITY_1, &xAudioHandle) != pdPASS) {
					printf("Failed to create APP task\r\n");
				}
			} else {
				printf("Cannot start audio - time not yet synced\r\n");
			}
			break;
		case 'm':
		case 'M':
			// Re-display Menu
			display_menu();
			break;
		}
	}
}

/**
 * Configure Console Function
 *
 * Configure the console UART.
 */
static void configure_console(void)
{
	const usart_serial_options_t uart_serial_options = {
		.baudrate = CONF_UART_BAUDRATE,
#if (defined CONF_UART_CHAR_LENGTH)
		.charlength = CONF_UART_CHAR_LENGTH,
#endif
		.paritytype = CONF_UART_PARITY,
#if (defined CONF_UART_STOP_BITS)
		.stopbits = CONF_UART_STOP_BITS,
#endif
	};

	/* Configure console UART. */
	stdio_serial_init(CONF_UART, &uart_serial_options);

	/* Specify that stdout should not be buffered. */
#if defined(__GNUC__)
	setbuf(stdout, NULL);
#else
	/* Already the case in IAR's Normal DLIB default configuration: printf()
	 * emits one character at a time.
	 */
#endif
}

/**
 * Allow Audio Request Function
 *
 * Re-enables the GPIO button (for debounce)
 */
static void allow_audio_request(void *parameter) {
	gpio_enable_pin_interrupt(PUSH_BUTTON_INT);
}

/**
 * PB0 Callback Function
 *
 * Push button's function - starts audio request
 */
static void pb0_callback(void) {
	gpio_disable_pin_interrupt(PUSH_BUTTON_INT);
	sw_timer_start(DEBOUNCE,500000,SW_TIMEOUT_RELATIVE,(FUNC_PTR)allow_audio_request,NULL);
	if (isTimeSynced() == 1) {
		if (xTaskCreate(audio_task, "AUDIO", TASK_1024_STACK_SIZE, NULL, TASK_STACK_PRIORITY_1, &xAudioHandle) != pdPASS) {
			printf("Failed to create APP task\r\n");
		}
	} else {
		printf("Cannot start audio - time not yet synced\r\n");
	}
}

/**
 * Main function
 */
int main(void)
{
	irq_initialize_vectors();
	sysclk_init();

	/* Initialize the board */
	board_init();

	sw_timer_init();
	
	/* Initialize the console uart */
	configure_console();

	/* Output demo infomation. */
	printf("-- GROUND NODE --\n\r");
	printf("-- %s\n\r", BOARD_NAME);
	printf("-- Compiled: %s %s --\n\r", __DATE__, __TIME__);
	
	if (MAC_SUCCESS != wpan_init()) {
		printf("No MAC SUCCESS\r\n");
	}

	/* Initialize LEDs. */
	LED_On(LED_START);     /* indicating application is started */
	LED_Off(LED_NWK_SETUP); /* indicating node is associated */
	LED_Off(LED_DATA);     /* indicating successfull data transmission */

	cpu_irq_enable();

	// Start USB device stack
	ui_init();
	ui_powerdown();
	udc_start();

	// Start timers
	sw_timer_get_id(&TIMER_LED_OFF);
	sw_timer_get_id(&DEBOUNCE);

	// Reset Node status
	node_status = COLD_START_RESET;

	/*
	 * Reset the MAC layer to the default values
	 * This request will cause a mlme reset confirm message ->
	 * usr_mlme_reset_conf
	 */
	wpan_mlme_reset_req(true);
	
	// Prepare DAC for audio output
	audio_init();

	// Setup Push Button
	ioport_set_pin_dir(PUSH_BUTTON_INT, IOPORT_DIR_INPUT);
	ioport_set_pin_mode(PUSH_BUTTON_INT, IOPORT_MODE_PULLUP | IOPORT_MODE_GLITCH_FILTER);
	ioport_set_pin_sense_mode(PUSH_BUTTON_INT, IOPORT_SENSE_FALLING);
	if (!gpio_set_pin_callback(PUSH_BUTTON_INT, pb0_callback, 1)) {
		printf("Set pin callback failure!\r\n");
		while(true){;}
	} else {
		gpio_enable_pin_interrupt(PUSH_BUTTON_INT);
	}

	/* Create WPAN task */
	if (xTaskCreate(wpan_run_task, "WPAN", TASK_2048_STACK_SIZE, NULL,
	TASK_STACK_PRIORITY_1, &xWpanHandle) != pdPASS) {
		printf("Failed to create WPAN task\r\n");
	}
	/* Create Console Task */
	if (xTaskCreate(console_task, "CONSOLE", TASK_1024_STACK_SIZE, NULL,
	TASK_STACK_PRIORITY_0, &xConsoleHandle) != pdPASS) {
		printf("Failed to create Console task\r\n");
	}
	/* Create USB Task (if needed) */
#ifdef   USB_DEVICE_LOW_SPEED
	if (xTaskCreate(low_speed_usb_task, "USB Low Speed", TASK_2048_STACK_SIZE, NULL,
	TASK_STACK_PRIORITY_0, &xUSBHandle) != pdPASS) {
		printf("Failed to create USB Task task\r\n");
	}
#endif

	/* Start the scheduler. */
	vTaskStartScheduler();

	/* Will only get here if there was insufficient memory to create the idle task. */
	return 0;
}

/*
 * Wireless Ready function
 *
 * Runs processes that can only be started after wireless is connected
 * Includes setting up CLOCK SYNC and AUDIO tasks
 */
void wireless_ready(void) {
	printf("Inside Wireless Ready\r\n");
	if (xClockSyncHandle == NULL) {
			if (xTaskCreate(clock_sync_task, "CLOCK", TASK_2048_STACK_SIZE, NULL,TASK_STACK_PRIORITY_0, &xClockSyncHandle) != pdPASS) {
				printf("Failed to create ADC task\r\n");
			}
	}
}

/**
 * Data Received function
 *
 * Handles any data that is received from the Zigbit module
 */
void data_received(uint8_t msduLength,uint8_t *msdu,uint32_t Timestamp) {
	static signed portBASE_TYPE xHigherPriorityTaskWoken;
	uint32_t sync_end_time = 0;
	uint32_t ack_start_time = 0;
	switch (msdu[0]) {
		case UNKNOWN:
			printf("Unknown data received\r\n");
			break;
		case ACK_PULSE:
			xSemaphoreGiveFromISR( xAckPulse, &xHigherPriorityTaskWoken );
			set_ack_end_time(Timestamp);
			break;
		case TIMESTAMP_PULSE:
			sync_end_time = ((msdu[1]*256+msdu[2])*256+msdu[3])*256+msdu[4];
			ack_start_time = ((msdu[5]*256+msdu[6])*256+msdu[7])*256+msdu[8];
			set_sync_end_time(sync_end_time);
			set_ack_start_time(ack_start_time);
			xSemaphoreGiveFromISR( xTimestampPulse, &xHigherPriorityTaskWoken );
			break;
		case RECORDING_STATUS:
			vTaskResume(xAudioHandle);
			break;
		default:
			printf("Bad Message Received\r\n");
			break;
	}
}

/**
 * Data Transmitted function
 *
 * Called by the WPAN task upon successful completion of a
 * sent message.  Other tasks may need to know when a message
 * has been properly sent.
 */
void data_transmitted(uint8_t msduHandle, uint8_t status, uint32_t Timestamp) {
	if (msduHandle == msduHandle_sync) {
		set_sync_start_time(Timestamp);
	}
	if (msduHandle == get_audio_start_handle()) {
		printf("Resuming Task\r\n");
		vTaskResume(xAudioHandle);
	}
}

