/**
 * File: time.c
 * Creator: Jonathan Klapel
 * Date: 08 June 2014
 * Description: Performs the synchronization and time-keeping
 */ 

#include "time.h"

/* Time variables */
static uint32_t sync_start = 0;
static uint32_t sync_end = 0;
static uint32_t ack_start = 0;
static uint32_t ack_end = 0;
static uint32_t prev_ack_end = 0;

/* Delta variables */
static int32_t delta = 0;
static int32_t transmit = 0;
static double delta_change = 0;

/* Synchronization boolean */
static uint8_t synced = 0;

/**
 * Set Sync Start Time Function
 *
 * Sets the sync start time for later use
 */
void set_sync_start_time(uint32_t time) {
	sync_start = time;
}

/**
 * Set Sync End Time Function
 *
 * Sets the sync end time for later use
 */
void set_sync_end_time(uint32_t time) {
	sync_end = time;
}

/**
 * Set ACK Start Time Function
 *
 * Sets the ACK start time for later use
 */
void set_ack_start_time(uint32_t time) {
	ack_start = time;
}

/**
 * Set ACK End Time Function
 *
 * Sets the ACK end time for later use
 */
void set_ack_end_time(uint32_t time) {
	prev_ack_end = ack_end;
	ack_end = time;
}

/**
 * Calculate Offsets Function
 *
 * Performs TPSN synchronization calculations and sets synchronization variables
 */
void calculate_offsets(void) {
	static uint8_t sync_count = 0;
	int32_t old_delta = delta;
	int32_t temp_delta = ((int32_t)(sync_end-sync_start)-(int32_t)(ack_end-ack_start))/2;
	
	if (old_delta == 0) {
		delta = temp_delta;
		delta_change = (float)(delta -old_delta)/(float)(ack_end-prev_ack_end);
		transmit = ((sync_end-sync_start)+(ack_end-ack_start))/2;
	} else if(abs(old_delta-temp_delta) < abs(delta_change*(ack_end-prev_ack_end))*1.5) {
		sync_count++;
		if (sync_count >= 2) {
			synced = 1;
			sync_count = 2;
		}
		delta = temp_delta;
		delta_change = (float)(delta -old_delta)/(float)(ack_end-prev_ack_end);
		transmit = ((sync_end-sync_start)+(ack_end-ack_start))/2;
	} else {
		printf("Missed delta max\r\n");
		synced = 0;
		ack_end = prev_ack_end;
		delta_change = delta_change*1.5;
	}
}

/**
 * Get Global Time Function
 *
 * Gets the current time on the coordinator node
 */
uint32_t get_global_time(void) {
	return (uint32_t)((int32_t)sw_timer_get_time()+delta+get_estimated_delta_change(sw_timer_get_time()));
}

/**
 * Get Delta Function
 *
 * Returns the most recently calculated delta
 */
int32_t get_delta(void) {
	return delta;
}

/**
 * Get Delta Change Function
 *
 * Returns the most recent delta change value
 */
float get_delta_change(void) {
	return delta_change;
}

/**
 * Get Estimated Delta Change Function
 *
 * Returns an estimate of how much the delta has changed since the last ACK timestamp
 */
int32_t get_estimated_delta_change(uint32_t timestamp) {
	return (int32_t)(delta_change*(float)(timestamp-ack_end));	
}

/**
 * Get Transmit Function
 *
 * Returns the most recent transmit time
 */
int32_t get_transmit(void) {
	return transmit;
}

/**
 * Is Time Synced Function
 *
 * Returns 1 if nodes are synchronized, otherwise 0
 */
uint8_t isTimeSynced(void) {
	return synced;
}