/**
 * File: ui.c
 * Creator: Jonathan Klapel
 * Date: 08 June 2014
 * Description: USB Interface methods
 */ 

#include <asf.h>
#include "ui.h"
#include "led.h"

static bool ui_b_led_blink = true;
static uint8_t ui_hid_report[UDI_HID_REPORT_OUT_SIZE];

/**
 * UI Init Function
 *
 * Initializes the user interface
 */
void ui_init(void)
{
	// Initialize LEDs
	LED_On(LED0);
}

/**
 * UI Powerdown Function
 *
 * Enters the user interface in power down mode
 */
void ui_powerdown(void)
{
	LED_Off(LED0);
}

/**
 * UI Wakeup Enable Function
 * 
 * Enables the asynchronous interrupts of the user interface
 */
void ui_wakeup_enable(void)
{
}

/**
 * UI Wakeup Disable Function
 * 
 * Disables the asynchronous interrupts of the user interface
 */
void ui_wakeup_disable(void)
{
}

/**
 * UI Wakeup Function
 *
 * Exits the user interface of power down mode
 */
void ui_wakeup(void)
{
	LED_On(LED0);
}

/**
 * UI Process Function
 *
 * This process is called each 1ms and only if the USB interface is enabled.
 */
void ui_process(void) {
	
	if (!main_b_generic_enable)
	return;
	uint16_t framenumber = udd_get_frame_number();
	
	bool b_btn_state;
	static bool btn0_last_state = false;
	static uint8_t cpt_sof = 0;

	// Blink LED
	if (ui_b_led_blink) {
		if ((framenumber % 1000) == 0) {
			LED_On(LED0);
		}
		if ((framenumber % 1000) == 500) {
			LED_Off(LED0);
		}
	}

	// Scan process running each 40ms
	cpt_sof++;
	if (cpt_sof < 40) {
		return;
	}
	cpt_sof = 0;

	// Scan buttons
	b_btn_state = !ioport_get_pin_level(GPIO_PUSH_BUTTON_0);
	if (b_btn_state != btn0_last_state) {
		ui_hid_report[0] = b_btn_state;
		udi_hid_generic_send_report_in(ui_hid_report);
		btn0_last_state = b_btn_state;
	}
}

/**
 * UI Report Out
 *
 * This function is called when a HID report is received
 */
void ui_report_out(uint8_t *report)
{
	uint64_t time;
	uint8_t payload[8];
	double fstart;
	double fend;
	double duration;
	double record_duration;
	xTaskHandle xAudioHandle;
	
	switch (report[0]) {
	case 0:
		// HELO Request
		time = ((uint64_t)report[7] * 256) + ((uint64_t)report[6] << 40) + ((uint64_t)report[5] << 32) + ((uint64_t)report[4] << 24) + ((uint64_t)report[3] << 16) + ((uint64_t)report[2] << 8) +  (uint64_t)report[1];
		printf("USB HELO Request - Time: %u\r\n", time);
		break;
	case 1:
		// Audio Request
		fstart = (double)(report[2]*256 + report[1]);
		fend = (double)(report[4]*256 + report[3]);
		duration = (double)report[5];
		record_duration = (double)report[6];
		set_playback_values(fstart,fend,duration,record_duration);
		if (isTimeSynced() == 1) {
			if (xTaskCreate(audio_task, "AUDIO", TASK_1024_STACK_SIZE*3, NULL, TASK_STACK_PRIORITY_1, &xAudioHandle) != pdPASS) {
				printf("Failed to create APP task\r\n");
			}
			ui_hid_report[1] = 1;
		} else {
			printf("Cannot start audio - time not yet synced\r\n");
			ui_hid_report[1] = 0;
		}
		ui_hid_report[0] = 2;
		udi_hid_generic_send_report_in(ui_hid_report);
		printf("USB Audio Request\r\n");
		break;
	case 2:
		// Audio Confirm - Should never receive here
		printf("USB Audio Confirm\r\n");
		break;
	case 3:
		// Connection Status Request
		ui_hid_report[0] = 3;
		ui_hid_report[1] = node_status;
		ui_hid_report[2] = isTimeSynced();
		udi_hid_generic_send_report_in(ui_hid_report);
		printf("USB Connection Status Request\r\n");
		break;
	}
}

/**
 * Main Generic Enable Function
 * 
 * Called by HID interface
 * Callback running when USB Host enable generic interface
 * 
 * Return: true if generic startup is ok
 */
bool main_generic_enable(void)
{
	main_b_generic_enable = true;
	return true;
}

/**
 * Main Generic Disable Function
 *
 * Called by HID interface
 * Callback running when USB Host disable generic interface
 */
void main_generic_disable(void)
{
	main_b_generic_enable = false;
}

/**
 * Main Hid Set Feature Function
 *
 * Called by UDI HID generic when USB Host send a feature request
 */
void main_hid_set_feature(uint8_t* report)
{
	if (report[0] == 0xAA && report[1] == 0x55 && report[2] == 0xAA && report[3] == 0x55) {
		// Disconnect USB Device
		udc_stop();
		ui_powerdown();
	}
}