/**
 * File: zigbit.c
 * Creator: Jonathan Klapel
 * Date: 08 June 2014
 * Description: Communicates over the Atmel Zigbit Module
 * Note: The original code for this page came from ASF's Zigbit example.
 *   Not all of the functions listed here are used, but must be included
 *   per the ASF ZigBit Library.
 */ 

#include "zigbit.h"

/**
 * Send Message Function
 *
 * Sends a message to the other node via Zigbit
 *
 * Takes the Message Type, the length of the message, and a pointer to the data
 */
uint8_t send_message(msg_type msgtype, uint8_t length, uint8_t * data) {
	node_status = TRANSMITTING;
	uint8_t src_addr_mode;
	wpan_addr_spec_t dst_addr;
	uint8_t payload[length+1];
	static uint8_t msduHandle = 0;
	
	src_addr_mode = WPAN_ADDRMODE_SHORT;
	dst_addr.AddrMode = WPAN_ADDRMODE_SHORT;
	dst_addr.PANId = DEFAULT_PAN_ID;
	
	payload[0] = msgtype;	
	for (uint8_t i = 0; i < length; i++) {
		payload[i+1] = data[i];
	}
	
	/* Perform different tasks depending on who this is */
#if (DEVICE == 1)
	ADDR_COPY_DST_SRC_16(dst_addr.Addr.short_address,coord_addr.short_addr);
#endif
#if (COORDINATOR == 1)
	ADDR_COPY_DST_SRC_16(dst_addr.Addr.short_address,device_list[0].short_addr);
#endif

	msduHandle++;
	
	/* Perform the actual send without interruption */
	taskENTER_CRITICAL();
	bool success = wpan_mcps_data_req(src_addr_mode,&dst_addr,length+1,&payload,msduHandle,WPAN_TXOPT_ACK);
	taskEXIT_CRITICAL();
	
	return msduHandle;
}

/**
 * User MCPS Data Confirm Function
 *
 * Called by Zigbit library on successful send
 */
#if defined(ENABLE_TSTAMP)
void usr_mcps_data_conf(uint8_t msduHandle,
		uint8_t status,
		uint32_t Timestamp)
#else
void usr_mcps_data_conf(uint8_t msduHandle,
		uint8_t status)
#endif
{
	node_status = IDLE;

	if (status == MAC_SUCCESS) {
		data_transmitted(msduHandle, status, Timestamp);
		LED_Off(LED_DATA);
	}

	/* Keep compiler happy. */
	msduHandle = msduHandle;
}

/**
 * User MCPS Data Indicate Function
 *
 * Called by Zigbit library on data reception
 */
void usr_mcps_data_ind(wpan_addr_spec_t *SrcAddrSpec,
		wpan_addr_spec_t *DstAddrSpec,
		uint8_t msduLength,
		uint8_t *msdu,
		uint8_t mpduLinkQuality,
#ifdef ENABLE_TSTAMP
		uint8_t DSN,
		uint32_t Timestamp)
#else
		uint8_t DSN)
#endif  /* ENABLE_TSTAMP */
{
	data_received(msduLength,msdu,Timestamp);
	
	LED_On(LED_DATA);
	
	/* Start a timer switching off the LED */
	sw_timer_start(TIMER_LED_OFF, 500000, SW_TIMEOUT_RELATIVE, (FUNC_PTR)data_exchange_led_off_cb, NULL);

	/* Keep compiler happy. */
	SrcAddrSpec = SrcAddrSpec;
	DstAddrSpec = DstAddrSpec;
	msduLength = msduLength;
	msdu = msdu;
	mpduLinkQuality = mpduLinkQuality;
	DSN = DSN;
}

#if ((MAC_PURGE_REQUEST_CONFIRM == 1) && (MAC_INDIRECT_DATA_BASIC == 1))
void usr_mcps_purge_conf(uint8_t msduHandle,
		uint8_t status)
{
}

#endif

/**
 * User MLME Associate Confirm Function
 *
 * Confirmation of correct association with another node
 */
#if (MAC_ASSOCIATION_REQUEST_CONFIRM == 1)
void usr_mlme_associate_conf(uint16_t AssocShortAddress,
		uint8_t status)
{
	if (status == MAC_SUCCESS) {
		sw_timer_stop(TIMER_LED_OFF);
		LED_On(LED_NWK_SETUP);
		node_status = IDLE;
		
		wireless_ready();
	} else {
		/* Something went wrong; restart */
		wpan_mlme_reset_req(true);
	}

	/* Keep compiler happy. */
	AssocShortAddress = AssocShortAddress;
}

#endif  /* (MAC_ASSOCIATION_REQUEST_CONFIRM == 1) */

/**
 * User MLME Associate Indicator
 *
 * Indicates when another node has successfully associated
 */
#if (MAC_ASSOCIATION_INDICATION_RESPONSE == 1)
void usr_mlme_associate_ind(uint64_t DeviceAddress,
		uint8_t CapabilityInformation)
{
	uint16_t associate_short_addr = macShortAddress_def;

	if (assign_new_short_addr(DeviceAddress,
			&associate_short_addr) == true) {
		wpan_mlme_associate_resp(DeviceAddress,
				associate_short_addr,
				ASSOCIATION_SUCCESSFUL);
	} else {
		wpan_mlme_associate_resp(DeviceAddress,
				associate_short_addr,
				PAN_AT_CAPACITY);
	}

	/* Keep compiler happy. */
	CapabilityInformation = CapabilityInformation;
}

#endif

#if (MAC_BEACON_NOTIFY_INDICATION == 1)
void usr_mlme_beacon_notify_ind(uint8_t BSN,
		wpan_pandescriptor_t *PANDescriptor,
		uint8_t PendAddrSpec,
		uint8_t *AddrList,
		uint8_t sduLength,
		uint8_t *sdu)
{
}

#endif

/**
 * User MLME Communication Status Indicator Function
 *
 * Gives the status after successful communication between nodes
 */
#if ((MAC_ORPHAN_INDICATION_RESPONSE == 1) || \
	(MAC_ASSOCIATION_INDICATION_RESPONSE == 1))
void usr_mlme_comm_status_ind(wpan_addr_spec_t *SrcAddrSpec,
		wpan_addr_spec_t *DstAddrSpec,
		uint8_t status)
{
	if (status == MAC_SUCCESS) {
		node_status = IDLE;
		wireless_ready();
	}

	/* Keep compiler happy. */
	SrcAddrSpec = SrcAddrSpec;
	DstAddrSpec = DstAddrSpec;
}

#endif

#if (MAC_DISASSOCIATION_BASIC_SUPPORT == 1)
void usr_mlme_disassociate_conf(uint8_t status,
		wpan_addr_spec_t *DeviceAddrSpec)
{
}

#endif

#if (MAC_DISASSOCIATION_BASIC_SUPPORT == 1)
void usr_mlme_disassociate_ind(uint64_t DeviceAddress,
		uint8_t DisassociateReason)
{
}

#endif

/**
 * User MLME Get Configuration
 *
 * Searches for a node to connect with
 */
#if (MAC_GET_SUPPORT == 1)
void usr_mlme_get_conf(uint8_t status,
		uint8_t PIBAttribute,
		void *PIBAttributeValue)
{
	if ((status == MAC_SUCCESS) && (PIBAttribute == phyCurrentPage)) {
		current_channel_page = *(uint8_t *)PIBAttributeValue;
		wpan_mlme_get_req(phyChannelsSupported);
	} else if ((status == MAC_SUCCESS) &&
			(PIBAttribute == phyChannelsSupported)) {
		uint8_t index;

		channels_supported = *(uint32_t *)PIBAttributeValue;

		for (index = 0; index < 32; index++) {
			if (channels_supported & (1 << index)) {
				current_channel = index + CHANNEL_OFFSET;
				break;
			}
		}

		wpan_mlme_scan_req(MLME_SCAN_TYPE_ACTIVE,
				SCAN_CHANNEL,
				SCAN_DURATION,
				current_channel_page);

		/* Indicate network scanning by a LED flashing */
		sw_timer_start(TIMER_LED_OFF,
				500000,
				SW_TIMEOUT_RELATIVE,
				(FUNC_PTR)network_scan_indication_cb,
				NULL);
	}
}

#endif

#if (MAC_ORPHAN_INDICATION_RESPONSE == 1)
void usr_mlme_orphan_ind(uint64_t OrphanAddress)
{
}

#endif

#if (MAC_INDIRECT_DATA_BASIC == 1)
void usr_mlme_poll_conf(uint8_t status)
{
}

#endif

void usr_mlme_reset_conf(uint8_t status)
{
	if (status == MAC_SUCCESS) {
		wpan_mlme_get_req(phyCurrentPage);
	} else {
		/* something went wrong; restart */
		wpan_mlme_reset_req(true);
	}
}

#if (MAC_RX_ENABLE_SUPPORT == 1)
void usr_mlme_rx_enable_conf(uint8_t status)
{
}

#endif  /* (MAC_RX_ENABLE_SUPPORT == 1) */

/**
 * User MLME Scan Configuration
 *
 * Scans for a broadcast network
 */
#if ((MAC_SCAN_ED_REQUEST_CONFIRM == 1)      ||	\
	(MAC_SCAN_ACTIVE_REQUEST_CONFIRM == 1)  || \
	(MAC_SCAN_PASSIVE_REQUEST_CONFIRM == 1) || \
	(MAC_SCAN_ORPHAN_REQUEST_CONFIRM == 1))
void usr_mlme_scan_conf(uint8_t status,
		uint8_t ScanType,
		uint8_t ChannelPage,
		uint32_t UnscannedChannels,
		uint8_t ResultListSize,
		void *ResultList) {
	if (status == MAC_SUCCESS) {
/* Different functions for each device */
#if (DEVICE == 1)
printf("Scanning for coordinator\r\n");
		wpan_pandescriptor_t *coordinator;
		uint8_t i;

		/*
		 * Analyze the Result List
		 * Assume that the first entry of the result list is our coodinator.
		 */
		coordinator = (wpan_pandescriptor_t *)ResultList;
		for (i = 0; i < ResultListSize; i++) {
			/*
			 * Check if the PAN descriptor belongs to our coordinator.
			 * Check if coordinator allows association.
			 */
			if ((coordinator->LogicalChannel == current_channel) &&
					(coordinator->ChannelPage == current_channel_page) &&
					(coordinator->CoordAddrSpec.PANId == DEFAULT_PAN_ID) &&
					((coordinator->SuperframeSpec &	((uint16_t)1 <<	ASSOC_PERMIT_BIT_POS)) == ((uint16_t)1 << ASSOC_PERMIT_BIT_POS))) {
						
				/* Store the coordinator's address */
				if (coordinator->CoordAddrSpec.AddrMode == WPAN_ADDRMODE_SHORT) {
					ADDR_COPY_DST_SRC_16(coord_addr.short_addr,coordinator->CoordAddrSpec.Addr.short_address);
				} else if (coordinator->CoordAddrSpec.AddrMode == WPAN_ADDRMODE_LONG) {
					ADDR_COPY_DST_SRC_64(coord_addr.ieee_addr,coordinator->CoordAddrSpec.Addr.long_address);
				} else {
					/* Something went wrong; restart */
					wpan_mlme_reset_req(true);
					return;
				}

				/* Associate to our coordinator */
				wpan_mlme_associate_req(coordinator->LogicalChannel, coordinator->ChannelPage, &(coordinator->CoordAddrSpec), WPAN_CAP_ALLOCADDRESS);
				return;
			}

			/* Get the next PAN descriptor */
			coordinator++;
		}

		/*
		 * If here, the result list does not contain our expected coordinator.
		 * Let's scan again.
		 */
		wpan_mlme_scan_req(MLME_SCAN_TYPE_ACTIVE, SCAN_CHANNEL, SCAN_DURATION, current_channel_page);
#endif		
#if (COORDINATOR == 1)
			/*
			 * Set the short address of this node.
			 */
			uint8_t short_addr[2];
			short_addr[0] = (uint8_t)COORD_SHORT_ADDR; /* low byte */
			short_addr[1] = (uint8_t)(COORD_SHORT_ADDR >> 8); /* high byte */
			wpan_mlme_set_req(macShortAddress, short_addr);
#endif
	} else if (status == MAC_NO_BEACON) {
#if (DEVICE == 1)
		/*
		 * No beacon is received; no coordiantor is located.
		 * Scan again, but used longer scan duration.
		 */
		wpan_mlme_scan_req(MLME_SCAN_TYPE_ACTIVE,SCAN_CHANNEL,SCAN_DURATION,current_channel_page);
#endif
#if (COORDINATOR == 1)
			uint8_t short_addr[2];
			short_addr[0] = (uint8_t)COORD_SHORT_ADDR; /* low byte */
			short_addr[1] = (uint8_t)(COORD_SHORT_ADDR >> 8); /* high byte */
			wpan_mlme_set_req(macShortAddress, short_addr);
#endif
	} else {
		/* Something went wrong; restart */
		wpan_mlme_reset_req(true);
	}

	/* Keep compiler happy. */
	ScanType = ScanType;
	UnscannedChannels = UnscannedChannels;
	ChannelPage = ChannelPage;
}

#endif

/** 
 * User MLME Set Configuration
 *
 * Set the device configuration
 */
void usr_mlme_set_conf(uint8_t status,uint8_t PIBAttribute) {
	if ((status == MAC_SUCCESS) && (PIBAttribute == macShortAddress)) {
		/*
		 * Allow other devices to associate to this coordinator.
		 */
		uint8_t association_permit = true;

		wpan_mlme_set_req(macAssociationPermit, &association_permit);
		
	} else if ((status == MAC_SUCCESS) &&
			(PIBAttribute == macAssociationPermit)) {
				
		bool rx_on_when_idle = true;

		wpan_mlme_set_req(macRxOnWhenIdle, &rx_on_when_idle);
	} else if ((status == MAC_SUCCESS) && (PIBAttribute == macRxOnWhenIdle)) {
		wpan_mlme_start_req(DEFAULT_PAN_ID, current_channel, current_channel_page, 15, 15, true, false, false);
	} else {
		/* something went wrong; restart */
		wpan_mlme_reset_req(true);
	}
}

/**
 * User MLME Start Configuration Function
 *
 * Set the node status to the full run mode
 */
#if (MAC_START_REQUEST_CONFIRM == 1)
void usr_mlme_start_conf(uint8_t status)
{
	if (status == MAC_SUCCESS) {
		node_status = PC_RUNNING;

		/* Stop timer used for search indication */
		sw_timer_stop(TIMER_LED_OFF);
		LED_On(LED_NWK_SETUP);
	} else {
		/* something went wrong; restart */
		wpan_mlme_reset_req(true);
	}
}

#endif  /* (MAC_START_REQUEST_CONFIRM == 1) */

void usr_mlme_sync_loss_ind(uint8_t LossReason,
		uint16_t PANId,
		uint8_t LogicalChannel,
		uint8_t ChannelPage)
{
}

/**
 * Assign New Short Address Function
 *
 * Assigns a new short address for an associating node
 */
static bool assign_new_short_addr(uint64_t addr64, uint16_t *addr16) {
	uint8_t i;

	/* Check if device has been associated before */
	for (i = 0; i < MAX_NUMBER_OF_DEVICES; i++) {
		if (device_list[i].short_addr == 0x0000) {
			/* If the short address is 0x0000, it has not been used
			 * before */
			continue;
		}

		if (device_list[i].ieee_addr == addr64) {
			/* Assign the previously assigned short address again */
			*addr16 = device_list[i].short_addr;
			return true;
		}
	}

	for (i = 0; i < MAX_NUMBER_OF_DEVICES; i++) {
		if (device_list[i].short_addr == 0x0000) {
			*addr16 = CPU_ENDIAN_TO_LE16(i + 0x0001);
			device_list[i].short_addr = CPU_ENDIAN_TO_LE16(i + 0x0001);
			device_list[i].ieee_addr = addr64;
			return true;
		}
	}

	/* If we are here, no short address could be assigned. */
	return false;
}

/**
 * Network Scan Indication Callback Function
 *
 * Toggles an LED while scanning for network
 */
static void network_scan_indication_cb(void *parameter)
{
	LED_Toggle(LED_NWK_SETUP);

	/* Re-start led timer again */
	sw_timer_start(TIMER_LED_OFF, 500000, SW_TIMEOUT_RELATIVE, (FUNC_PTR)network_scan_indication_cb, NULL);

	parameter = parameter; /* Keep compiler happy. */
}

/**
 * Data Exchange LED Off Callback Function
 *
 * Disables the LED after data has been transferred
 */
static void data_exchange_led_off_cb(void *parameter)
{
	LED_Off(LED_DATA);

	parameter = parameter; /* Keep compiler happy. */
}
