/**
* File: main.cpp
* Creator: Jonathan Klapel
* Date: 08 June 2014
* Description: USB sample application
*
* APPLICATION USE
* Use:
* - Using the USB cable, connect the SAM4L USB port to the PC
* - The PC should automatically install a generic HID driver
* - Run the application and follow the console menu
*
* Notes:
*  Simple code examples are given here to demonstrate how to use the USB
*  library.  For more details on the Atmel AtUsbHid driver, see
*  http://www.atmel.com/dyn/resources/prod_documents/doc7645.pdf
*  and http://www.atmel.com/dyn/resources/prod_documents/AVR153.zip
*/

#include <iostream>
#include <stdio.h>
#include <windows.h>
#include <thread>
#include "AtUsbHid.h"

using namespace std;

// USB HID device Vendor ID and Product ID.
#define VID 0x03EB
#define PID_1 0x2402
#define PID_2 0x2013

// USB message types
#define MSG_HELO 0
#define MSG_AUDIO_REQUEST 1
#define MSG_AUDIO_CONFIRM 2
#define MSG_STATUS 3

bool running = false;

/**
 * USB Reads Function
 *
 * Reads from the USB and reacts to different message types
 */
void usb_reads(void) {
	unsigned char sbuffer[256];
	// Continue reading until application exit
	while (running) {
		// Read from the USB
		if (DYNCALL(readData(sbuffer)) != 0) {
			// React to message
			switch (sbuffer[0]) {
			case MSG_HELO:
				cout << "Received HELO Request" << endl;
				break;
			case MSG_AUDIO_REQUEST:
				cout << "Received Audio Request" << endl;
				break;
			case MSG_AUDIO_CONFIRM:
				cout << "Received Audio Confirm" << endl;
				break;
			case MSG_STATUS:
				cout << "Received Connection Status" << endl;
				if (sbuffer[1] == 2 || sbuffer[1] == 3)
					cout << "Wireless Connection Established" << endl;
				else
					cout << "Wireless Connection Not Established" << endl;
				if (sbuffer[2] == 1)
					cout << "Nodes Synced" << endl;
				else
					cout << "Nodes Not Synced" << endl;
				break;
			}
		}
	}
}

int main(void);
/**
 * Main Function
 *
 * Starting function for the application.  Sets up and carries out USB calls
 */
int main(void) {
	// Write Buufer used to set data_out
	unsigned char	data_out[9];

	// Handle to AtUsbHid.dll
	HINSTANCE hLib = NULL;  // Handle to our USB HID DLL.

	printf("Atmel USB HID Library Test Program\n\n");

	// Explicitely load the AtUsbHid library.
	printf("Loading USB HID Dll.\n");
	hLib = LoadLibraryA(AT_USB_HID_DLL);
	if (hLib != NULL) {
		printf("USB HID Dll loaded\n");
	} else {
		if (GetLastError() == ERROR_MOD_NOT_FOUND) {
			printf("Error: Could not find the Atmel USB HID Dll.\n");
			printf("       Please update the PATH variable.\n");
		} else {
			printf("Error: While opening Dll.\n");
		}
		return 1;
	}

	// Get USB HID library functions addresses.
	printf("Loading all Dll functions.\n");
	if (loadFuncPointers(hLib)) {
		printf("All function of the Dll has been loaded\n");
	}
	else {
		printf("Error: Could not find load function of the Dll.\n");
		return 1;
	}

	// Open our USB HID device.
	printf("Opening USB HID device with Vendor ID= 0x%04X and Product ID=0x%04X\n", VID, PID_1);
	if (DYNCALL(findHidDevice)(VID, PID_1)) {
		printf("USB HID device  VID=0x%04X, PID=0x%04X opened.\n", VID, PID_1);
	}
	else {
		switch (GetLastError()) {
		case ERROR_USB_DEVICE_NOT_FOUND:
			printf("Error: Could not open the device.\n");
			break;
		case ERROR_USB_DEVICE_NO_CAPABILITIES:
			printf("Error: Could not get USB device capabilities.\n");
			break;
		default:
			printf("Error: While opening device.\n");
			break;
		}
	}

	// Print Some information about the connected device

	printf("USB HID Input   Buffer size is %dByte.\n", DYNCALL(getInputReportLength()));
	printf("USB HID Output  Buffer size is %dByte.\n", DYNCALL(getOutputReportLength()));
	printf("USB HID Feature Buffer size is %dByte.\n", DYNCALL(getFeatureReportLength()));


	
	// Start USB Read thread
	running = true;

	thread t1(usb_reads);
	int menu_choice = 0;

	// Display menu and react to choices.  Exit on "4"
	while (menu_choice != 4) {

		// Output Menu
		cout << endl << "Menu options" << endl;
		cout << "0: HELO Request" << endl;
		cout << "1: Audio Request" << endl;
		cout << "2: Audio Confirm" << endl;
		cout << "3: Connection Status Request" << endl;
		cout << "4: Exit" << endl;
		cout << "Enter menu number choice: ";

		// Read in menu selection choice
		cin >> menu_choice;

		LARGE_INTEGER lg;
		SYSTEMTIME st;

		// React to choice
		switch (menu_choice) {
			// 0: HELO Request - Send current system time
			case MSG_HELO:
				data_out[0] = 0;
				GetSystemTime(&st);
				lg.QuadPart = ((st.wHour * 60 + st.wMinute) * 60 + st.wSecond) * 1000 + st.wMilliseconds;
				data_out[1] = (lg.QuadPart >> 0) & 0xff;
				data_out[2] = (lg.QuadPart >> 8) & 0xff;
				data_out[3] = (lg.QuadPart >> 16) & 0xff;
				data_out[4] = (lg.QuadPart >> 24) & 0xff;
				data_out[5] = (lg.QuadPart >> 32) & 0xff;
				data_out[6] = (lg.QuadPart >> 40) & 0xff;
				data_out[7] = (lg.QuadPart >> 48) & 0xff;
				DYNCALL(writeData)(data_out);
				break;
			// 1: Audio Request - Request an audio measurement
			case MSG_AUDIO_REQUEST:
				data_out[0] = 1;
				data_out[1] = (500 >> 0) & 0xff;
				data_out[2] = (500 >> 8) & 0xff;
				data_out[3] = (10000 >> 0) & 0xff;
				data_out[4] = (10000 >> 8) & 0xff;
				data_out[5] = 5;
				data_out[6] = 5;
				DYNCALL(writeData)(data_out);
				break;
			// 2: Audio Confirm - Request confirmation of audio
			case MSG_AUDIO_CONFIRM:
				data_out[0] = 2;
				DYNCALL(writeData)(data_out);
				break;
			// 3: Connection Status Request - Request current node status
			case MSG_STATUS:
				data_out[0] = 3;
				DYNCALL(writeData)(data_out);
				break;
		}
	}

	// Finished application, tell read thread to stop
	running = false;
	t1.join();

	// Close our USB HID device.	
	DYNCALL(closeDevice)();
	printf(">>> USB HID device VID=0x%04X closed.\n", VID);

	// Wait thread to be close before freeing DLL
	Sleep(1000);
	FreeLibrary(hLib);
	printf(">>> Please press a key to exit");
	getchar();

	return 1;
}