function [ output_args ] = ir_response_compare(lsp_file,mic_file1,mic_file2,window_size,plot_title,smoothing_constant,offset)
%ir_response Computes and displays the impulse response of a recording
%
% Parameters:
%  lsp_file - Loudspeaker file (WAV)
%  mic_file1 - First microphone recording (WAV)
%  mic_file2 - Second microphone recording (WAV)
%  window_size - IR window size
%  plot_title - Title of the graph
%  smoothing_constant - Smoothing constant to use when graphing
%  offset - An offset to apply to mic_file2 when plotting

%Read input files
[sig,fs] = wavread(lsp_file);
[micrec1, fs_rec1] = wavread(mic_file1);
[micrec2, fs_rec2] = wavread(mic_file2);

%Check that Sampling Frequencies are equal
if fs ~= fs_rec1
    error('Sample Frequency Mismatch');
end
if fs ~= fs_rec2
    error('Sample Frequency Mismatch');
end

%Make sure we are only using mono files
sig = sig(:,1);
micrec1 = micrec1(:,1);
micrec2 = micrec2(:,1);

%Add default values
if nargin < 6
    smoothing_constant = 1;
end
if nargin < 5
    plot_title = mic_file1;
end
if nargin < 4
    window_size = size(micrec1,1);
end

%Begin analysis
nfft_analyze = 2^20;
TF_system1 = fft(micrec1,nfft_analyze)./fft(sig,nfft_analyze);
TF_system2 = fft(micrec2,nfft_analyze)./fft(sig,nfft_analyze);
ir_system1 = real(ifft(TF_system1));
ir_system2 = real(ifft(TF_system2));
ivdirectsoundwindow = [1:window_size];
ir_useful1 = ir_system1(ivdirectsoundwindow);
ir_useful2 = ir_system2(ivdirectsoundwindow);

%Plot Time Domain IR
figure(1)
plot(ir_useful1)
hold on
plot(ir_useful2,'r')
title(strcat('Time Domain IR - ',plot_title));
xlabel('Sample');
ylabel('Response');

%Plot Frequency Domain IR
nfft_plot = 16384;
TF_useful1 = fft(ir_useful1,nfft_plot);
TF_useful2 = fft(ir_useful2,nfft_plot);
fvec = fs/nfft_plot*[0:nfft_plot/2-1];
figure(2)
semilogx(fvec,smooth(20*log10(abs(TF_useful1(1:nfft_plot/2))),1))
hold on
semilogx(fvec,smooth(20*log10(abs(TF_useful2(1:nfft_plot/2)))+offset,1),'r')
title(strcat('Frequency Domain IR - ',plot_title));
xlabel('Frequency (Hz)');
ylabel('Response (dB scale)');
axis([0 100000 -100 10]);

figure(3)
semilogx(fvec,smooth(20*log10(abs(TF_useful1(1:nfft_plot/2))),smoothing_constant))
hold on
semilogx(fvec,smooth(20*log10(abs(TF_useful2(1:nfft_plot/2)))+offset,smoothing_constant),'r')
title(strcat('Frequency Domain IR - ',plot_title,'(smoothed)'));
xlabel('Frequency (Hz)');
ylabel('Response (dB scale)');
axis([0 100000 -100 10]);

%Save images
%saveas(1,strcat('Time Domain IR - ',plot_title,'.png'));
%saveas(2,strcat('Frequency Domain IR - ',plot_title,'.png'));
%saveas(3,strcat('Frequency Domain IR - ',plot_title,'(smoothed).png'));

end

